# (c) Copyright 2009-2011. CodeWeavers, Inc.

import errno
import fcntl
import os
import re
import stat
import subprocess
import sys
import time
import traceback

import cxlog
import distversion
import cxobjc

import gettext

class CXUtils(cxobjc.Proxy):
    pass

#####
#
# CX_ROOT
#
#####

# Points to CrossOver's root directory.
# This should be set at startup and is then considered to be constant for the
# duration of the process.
CX_ROOT = None



#####
#
# File and directory helpers
#
#####

def mkdirs(path, *args, **kwargs):
    """Similar to os.makedirs() but returns False to signal errors and logs
    errors.
    """
    try:
        os.makedirs(path, *args, **kwargs)
    except OSError:
        ose = sys.exc_info()[1]
        if ose.errno != errno.EEXIST:
            cxlog.err("unable to create the '%s' directory: %s" % (cxlog.debug_str(path), ose.strerror))
            return False
    return True


#####
#
# Cross-process lock
#
#####

class FLock(object):

    def __init__(self, name):
        """Creates and takes a lock called name using flock().

        This is used to avoid races between processes.
        """
        self._name = name

        # Create the lock directory and perform some sanity checks
        lockdir = os.path.join(os.environ.get('TMPDIR', '/tmp'),
                               ".wine-%d" % os.geteuid())
        mkdirs(lockdir, int('700', 8))
        st = os.lstat(lockdir)
        if not stat.S_ISDIR(st.st_mode):
            raise ValueError("'%s' is not owned by you" % lockdir)
        if st.st_uid != os.geteuid():
            raise ValueError("'%s' is not owned by you" % lockdir)
        if st.st_mode & int('77', 8):
            raise ValueError("'%s' must not be accessible by other users" % lockdir)

        # Perform some more sanity checks and create the lock file
        self._path = os.path.join(lockdir, name + ".lock")
        try:
            st = os.lstat(self._path)
            if not stat.S_ISREG(st.st_mode):
                raise ValueError("'%s' is not a regular file" % self._path)
        except OSError:
            ose = sys.exc_info()[1]
            if ose.errno != errno.ENOENT:
                raise ose
        self._lock = open(self._path, 'w')

        # Take the lock
        cxlog.log("%d: Grabbing the '%s' lock" % (os.getpid(), self._path))
        try:
            fcntl.flock(self._lock, fcntl.LOCK_EX)
        except Exception:
            e = sys.exc_info()[1]
            # The filesystem probably does not support locking
            cxlog.warn("unable to take the '%s' lock: %s" % (name, e))
            self._lock.close()
            self._lock = None
            raise # re-raise the exception

    def _getname(self):
        """Returns the lock name."""
        return self._name

    name = property(_getname)

    def locked(self):
        """Returns True if the lock has not been released yet."""
        return self._lock is not None

    def release(self):
        """Releases the lock.

        Note that deleting a lock carries a very high risk of causing races
        outside the scope of this class and has essentially no benefit. So the
        file created for the lock is not deleted and no support is provided
        for deleting it.
        """
        if self._lock:
            cxlog.log("%d: Releasing the '%s' lock" % (os.getpid(), self._name))
            try:
                try:
                    fcntl.flock(self._lock, fcntl.LOCK_UN)
                except Exception:
                    e = sys.exc_info()[1]
                    cxlog.warn("unable to release the '%s' lock: %s" % (self._name, e))
            finally:
                self._lock.close()
                self._lock = None


#####
#
# String handling
#
#####

try:
    # pylint: disable-msg=E0602
    bytes_type = bytes
except NameError:
    bytes_type = str

try:
    unicode_type = unicode
except NameError:
    unicode_type = str

def string_to_unicode(string):
    if isinstance(string, unicode_type):
        return string
    elif isinstance(string, bytes_type):
        return string.decode('utf8')
    else:
        raise ValueError("string_to_unicode called with non-string")

def expect_unicode(string):
    if isinstance(string, unicode_type):
        return string
    elif isinstance(string, bytes_type):
        cxlog.warn("Byte string passed where unicode string was expected:\n%s" % ''.join(traceback.format_stack()[0:-1]))
        return string.decode('utf8')
    else:
        raise ValueError("expect_unicode() called with non-string")

def unescape_string(string):
    """Unescapes a string that was placed in double-quotes."""
    string = string.replace('\\\"', '\"')
    string = string.replace("\\\\", "\\")
    return string


def escape_string(string):
    """Escapes the string so it can be placed in double-quotes."""
    string = string.replace("\\", "\\\\")
    string = string.replace('"', '\\"')
    return string


def shquote_string(string):
    """Quote a string so it can be used in shell commands.

    Note that this implies escaping '$'s and '`'s which may not be appropriate
    in another context.
    """
    string = string.replace('\\', '\\\\')
    string = string.replace('\$', '\\\$')
    string = string.replace('\"', '\\\"')
    string = string.replace("\'", "\\\'")
    return '"' + string + '"'

_BAD_SHELL_CHARS = re.compile("[^a-zA-Z0-9\/.,+_-]")
def maybe_shquote_string(string):
    if _BAD_SHELL_CHARS.search(string):
        return shquote_string(string)
    return string

def unmangle(string):
    result = []
    x = 0
    while x < len(string):
        char = string[x]
        x += 1
        if char == '+':
            result.append(' ')
        elif char == '_':
            result.append('/')
        elif char == '~':
            result.append('\\')
        elif char == '^':
            try:
                result.append(chr(int(string[x:x+2], 16)))
            except ValueError:
                # not followed by valid hex digits
                pass
            else:
                x += 2
        else:
            result.append(char)
    return ''.join(result)


_ACCELERATOR_RE = re.compile('&(.)')

def remove_accelerators(string):
    """Remove the ampersands that identify the accelerator keys on Windows.

    While only the first one is meant to be significant, they should all be
    removed, and '&&' should be turned into a regular '&'.
    """
    return _ACCELERATOR_RE.sub('\\1', string)


def md5_hasher():
    """Returns an object for computing MD5 hashes."""
    try:
        # pylint: disable-msg=F0401,E1101
        import hashlib
        return hashlib.md5()
    except ImportError:
        import md5
        return md5.new()


def bottlename_character_filter(character):
    """Returns False for characters that are best avoided
    in bottle names.
    """
    # \/*?:|<>" are not valid characters in Windows paths
    # {} are not valid in paths stored in Debian and Solaris packages
    return character not in "\\/*?:|<>\"'{}"


@cxobjc.method(CXUtils, 'sanitizedBottleName_')
def bottlename_sanitize(bottlename):
    """Removes characters that are best avoided
      in bottle names. """

    result = []
    for x in bottlename:
        if bottlename_character_filter(x):
            result.append(x)
        else:
            result.append('_')

    # . causes the bottle to be hidden if it's the first character
    return ''.join(result).lstrip(".")

def html_escape(string):
    """Escape characters that have special meaning in html."""
    replacements = {
        '<': '&lt;',
        '>': '&gt;',
        '&': '&amp;'}
    result = []
    for char in string:
        result.append(replacements.get(char, char))
    return ''.join(result)

_HTML_BR_RE = None
_HTML_P_RE = None
_HTML_LI_RE = None
_HTML_A_RE = None
_HTML_URL_RE = None
_HTML_TAGS_RE = None
_LF_RE = None
def html_to_text(string):
    """Use some heuristics to turn an HTML string into a readable plain
    text string."""
    if string is None:
        return None

    # pylint: disable-msg=W0603
    global _HTML_BR_RE, _HTML_P_RE, _HTML_LI_RE, _HTML_A_RE, _HTML_URL_RE
    global _HTML_TAGS_RE, _LF_RE
    if not _HTML_BR_RE:
        _HTML_BR_RE = re.compile('<br/?>(?:\s*\n)*', re.IGNORECASE)
        _HTML_P_RE = re.compile(r'''</?(?:div|p|ol|ul)(\s+[a-z]+=(['"])(?:[^\\'"]|\.)*\2)*>(?:\s*\n)*''', re.IGNORECASE)
        _HTML_LI_RE = re.compile(r'''(\s*\n)*\s*<li(\s+[a-z]+=(['"])(?:[^\\'"]|\\.)*\3)*>''', re.IGNORECASE)
        _HTML_A_RE = re.compile(r'''</?a(\s+[a-z]+=(['"])(?:[^\\'"]|\.)*\2)*>(?:\s*\n)*''', re.IGNORECASE)
        _HTML_URL_RE = re.compile(r'''\shref=(['"])((?:[^\\'"]|\.)*)\1''', re.IGNORECASE)
        _HTML_TAGS_RE = re.compile(r'''(?:</?[a-z]+(\s+[a-z]+=(['"])(?:[^\\'"]|\\.)*\2)*/?>)+''', re.IGNORECASE)
        _LF_RE = re.compile('\s*\n\s*\n(?:\s*\n)+')

    # Remove carriage returns
    string = string.replace('\r', '')

    # Some tags deserve special treatment
    string = _HTML_BR_RE.sub('\n', string)    # End of line tag
    string = _HTML_LI_RE.sub('\n * ', string) # Enum tag
    string = _HTML_P_RE.sub('\n\n', string)   # 'Paragraph' tags

    # Preserve the URL for links so users can copy/paste them to their browser
    def replace_link(match):
        atag = match.group(0)
        url = _HTML_URL_RE.search(atag)
        if url:
            return "[" + url.group(2) + "] "
        return ""
    string = _HTML_A_RE.sub(replace_link, string)

    # Simply remove all other tags
    string = _HTML_TAGS_RE.sub('', string)

    # Replace the most common entities
    string = string.replace('&lt;', '<')
    string = string.replace('&gt;', '>')
    string = string.replace('&amp;', '&')
    string = string.replace('&nbsp;', ' ')

    # Finally remove excess line feeds
    string = string.replace('\r\n', '\n')
    return _LF_RE.sub('\n\n', string).strip('\n')

@cxobjc.method(CXUtils, 'cmdLineToArgv_')
def cmdlinetoargv(string):
    """Convert a windows-style quoted command line to an argument list. This is
    NOT equivalent to CommandLineToArgvW."""
    in_quotes = False
    pos = 0
    result = []
    arg = []
    while pos < len(string):
        char = string[pos]
        if char == ' ' and not in_quotes:
            if arg:
                result.append(''.join(arg))
                del arg[:]
        elif char == '\\' and pos+1 < len(string) and string[pos+1] in '\\"':
            pos += 1
            arg.append(string[pos])
        elif char == '"':
            in_quotes = not in_quotes
        else:
            arg.append(string[pos])
        pos += 1
    if arg:
        result.append(''.join(arg))
    return result

def argvtocmdline(argv):
    """Convert an argument list to a windows-style quoted command line."""
    result = []
    for arg in argv:
        arg = arg.replace("\\", "\\\\").replace('"', '\\"')
        if ' ' in arg:
            arg = '"%s"' % arg
        result.append(arg)
    return ' '.join(result)

@cxobjc.method(CXUtils, 'basename_')
def basename(path):
    """Get the base name of a Windows or Unix path.
    Note that Windows paths can mix forward and backward slashes.
    """
    index = max(path.rfind('/'), path.rfind('\\'))
    if index >= 0:
        return path[index+1:]
    return path

@cxobjc.method(CXUtils, 'dirname_')
def dirname(path):
    """Get the directory name of a Windows or Unix path.
    Note that Windows paths can mix forward and backward slashes.
    """
    index = max(path.rfind('/'), path.rfind('\\'))
    if index >= 0:
        return path[:index].rstrip('/\\')
    return ''


_regex_specials = set('.^$*+?[]|()')

def unescape_regex(string):
    """If a regex can match multiple strings, return None. If it can match only
    one string, this function MAY return that string."""
    in_curly_brace = False
    escape = False
    result = []
    string = string.lstrip('^').rstrip('$')
    for ch in string:
        if ch == '{' and not escape:
            in_curly_brace = True
        elif in_curly_brace:
            if ch == '}':
                return None
            elif not ch.isdigit() and ch != ',':
                in_curly_brace = False

        if escape:
            if ch.isalnum():
                # An escaped letter or number might have special meaning
                return None
            result.append(ch)
            escape = False
        elif ch in _regex_specials:
            return None
        elif ch == '\\':
            escape = True
        else:
            result.append(ch)
    return ''.join(result)

_OCTALCHAR_RE = None
def expand_octal_chars(string):
    """Expand escaped octal characters of the form '\040' in the specified
    string.
    """
    if string is None:
        return None
    # pylint: disable-msg=W0603
    global _OCTALCHAR_RE
    if not _OCTALCHAR_RE:
        _OCTALCHAR_RE = re.compile('(\\\\[0-7]{3})')
    i = 0
    expanded = list()
    for match in _OCTALCHAR_RE.finditer(string):
        expanded.append(string[i:match.start()])
        expanded.append(chr(int(string[match.start()+1:match.end()], 8)))
        i = match.end()
    expanded.append(string[i:])
    return ''.join(expanded)

_UNIXVAR_RE = None
def expand_unix_string(environ, string):
    """Expand references to environment variables of the form '${VARNAME}' in
    the specified string.

    Note that unlike os.path.expandvars(), if a given environment variable is
    not set it expands to an empty string.
    """
    if string is None:
        return None
    # pylint: disable-msg=W0603
    global _UNIXVAR_RE
    if not _UNIXVAR_RE:
        _UNIXVAR_RE = re.compile('\${(\w+)}')
    i = 0
    expanded = list()
    for match in _UNIXVAR_RE.finditer(string):
        expanded.append(string[i:match.start()])
        if match.group(1) in environ:
            expanded.append(environ[match.group(1)])
        i = match.end()
    expanded.append(string[i:])
    return ''.join(expanded)

def _int_for_version_component(comp):
    """A version component can have a non-numeric suffix (e.g. the "local" in
    "25889local").  That interferes with converting to integer. This function
    collects just the numeric prefix and converts that.
    """
    s = ""
    for x in comp:
        if x.isdigit():
            s = s + x
        else:
            break
    return int(s)

def subtract_version(a, b):
    a = a.split('.') if a else ()
    b = b.split('.') if b else ()
    if len(a) < len(b):
        a.extend(["0"] * (len(b) - len(a)))
    elif len(b) < len(a):
        b.extend(["0"] * (len(a) - len(b)))

    result = []
    sign = 0
    for i in range(len(a)):
        if sign == 0:
            sign = cmp(_int_for_version_component(a[i]), _int_for_version_component(b[i]))
        result.append((_int_for_version_component(a[i]) - _int_for_version_component(b[i])) * sign)
    return tuple(result)

def cmp_versions(a, b):
    a = a.split('.') if a else ()
    b = b.split('.') if b else ()
    if len(a) < len(b):
        a.extend(["0"] * (len(b) - len(a)))
    elif len(b) < len(a):
        b.extend(["0"] * (len(a) - len(b)))

    for i in range(len(a)):
        result = cmp(_int_for_version_component(a[i]), _int_for_version_component(b[i]))
        if result != 0:
            return result
    return 0


#####
#
# Locale functions
#
#####

_SINGLETON_RE = re.compile("(-x)?-[a-z]$")
_PREFERRED_LANGUAGES = None
def get_preferred_languages(language=None):
    """Returns a sequence of the user's preferred languages if language is
    None; and a sequence of variants otherwise.

    All languages must be identified by their RFC 1766 / RFC 3066 / RFC 5646 /
    BCP 47 language id.
    In the None case the result corresponds to the user's preferences on Mac
    OS X, while on Unix it is derived from the user's locale.
    When starting from a single language, the variants added allow broader
    matches to be established. For instance 'pt-br' gives ('pt-br', 'pt', '').
    """
    # pylint: disable-msg=W0603
    global _PREFERRED_LANGUAGES
    if _PREFERRED_LANGUAGES is not None:
        return _PREFERRED_LANGUAGES

    if language:
        languages = (language,)
    else:
        languages = distversion.get_user_languages()
        if not languages:
            languages = ('en',)

    # Add the user's preferred languages
    seen = set()
    _PREFERRED_LANGUAGES = []
    fallbacks = []
    for language in languages:
        language = language.lower()
        if language in seen:
            continue
        _PREFERRED_LANGUAGES.append(language)
        seen.add(language)

        # Compute the fallbacks
        while '-' in language:
            language = language.rsplit('-', 1)[0]
            # Strip trailing singleton subtags such as '-x' as they have no
            # meaning on their own. See 'Extension Subtags' in RFC 5646 and
            # the lookup algorithm in RFC 4647.
            language = _SINGLETON_RE.sub('', language)
            if language not in seen:
                fallbacks.append((language, len(_PREFERRED_LANGUAGES)))

    # Take extra care to insert the fallbacks where they are useful and yet
    # don't override the user's preferences. Here are a few scenarios:
    #   ('en-us, 'fr')       -> ('en-us, 'en', 'fr')
    #   ('en-us, 'fr', 'en') -> ('en-us, 'fr', 'en')
    #   ('en-us, 'en-ca')    -> ('en-us, 'en-ca', 'en')
    #   ('zh-hant-tw', 'fr', 'zh-tw') -> ('zh-hant-tw', 'zh-hant', 'fr', 'zh-tw', 'zh')
    #   ('en-a-bbb-x-a-ccc-ddd') -> ('en-a-bbb-x-a-ccc-ddd', 'en-a-bbb-x-a-ccc', 'en-a-bbb', 'en')
    # See also the lookup algorithm in RFC 4647
    # Insert the fallbacks starting from the end of the array to not mess up
    # the first indices.
    for fallback in reversed(fallbacks):
        if fallback[0] in seen:
            continue
        _PREFERRED_LANGUAGES.insert(fallback[1], fallback[0])
        seen.add(fallback[0])

    # Add the ultimate fallback
    _PREFERRED_LANGUAGES.append('')
    return _PREFERRED_LANGUAGES


@cxobjc.method(CXUtils, 'getLanguageValue_forLanguage_')
def get_language_value(dictionary, language=None):
    """Returns the appropriate value for the specified language from a
    dictionary mapping language names to values.

    The given dictionary should either contain a single '' key that is used
    for all languages, or only have locales as its keys for locale-specific
    values. However for compatibility with pre-10.0 releases, dictionaries
    containing both types of keys are supported too.
    Empty values are ignored.

    This function will always return a non-empty value if there is at least one
    in the dictionary.
    """
    languages = get_preferred_languages(language)
    for lang in languages:
        if lang in dictionary and dictionary[lang]:
            return (lang, dictionary[lang])

    if dictionary:
        # The dictionary is not empty, but there is no default. Just pick
        # something, predictably.
        languages = dictionary.keys()
        languages.sort()
        for lang in languages:
            if dictionary[lang]:
                return (lang, dictionary[lang])
    return ('', '')


_gettext_translator = None

def cxgettext(message):
    # pylint: disable-msg=W0603
    global _gettext_translator
    if not _gettext_translator:
        languages = []
        for lang in get_preferred_languages():
            # We must convert the language id to match the naming of the
            # $CX_ROOT/share/locale folders. It's not case sensitive, but the
            # dashes must be replaced by underscores.
            languages.append(lang.replace('-', '_'))
        locale_path = os.path.join(CX_ROOT, "share", "locale")
        cxlog.log("locale path %s" % cxlog.debug_str(locale_path))
        gettext.bindtextdomain("crossover", locale_path)
        gettext.bind_textdomain_codeset("crossover", "UTF-8")
        gettext.textdomain("crossover")
        _gettext_translator = gettext.translation("crossover", locale_path, languages=languages, fallback=True)
    return string_to_unicode(_gettext_translator.gettext(message))


#####
#
# Dealing with subprocesses
#
#####

GRAB = 1
NULL = 2
STDOUT = 3
NOLOG = 4

def run(args, bufsize=0, executable=None, stdout=None, stderr=None,
        preexec_fn=None, shell=False, background=False, cwd=None, env=None,
        universal_newlines=True, logprefix=""):
    """Works the same way as subprocess.Popen, but simplifies retrieving the
    output and exit code. Returns (retcode, out, err) triplets.

    If stdout or stderr are set to GRAB, then they are captured and returned.
    If they are set to NULL, then they are redirected to /dev/null instead.
    If stderr is set to STDOUT, then it is redirected to stdout.
    If the command cannot be run, then returns -1 as the exit code instead of
    raising an exception.

    Finally, the execution is timed and traced with cxlog. If stdout or stderr
    are being captured, then they are logged too, but that can be disabled by
    or-ing NOLOG. If logprefix is set, then the trace is prefixed with it.
    """
    if cxlog.is_on():
        start = time.time()
        if logprefix != "":
            logprefix = logprefix + ": "
        bg = ""
        if background:
            bg = " &"
        cxlog.log("%sRunning: %s%s" % (logprefix, " ".join(cxlog.debug_str(arg) for arg in args), bg))

    out_log = False
    if stdout:
        if (stdout & ~NOLOG) == GRAB and not background:
            subout = subprocess.PIPE
            if stdout == GRAB:
                out_log = True
        elif stdout == NULL:
            subout = open("/dev/null", "w")
        else:
            raise TypeError("invalid value for stdout")
    else:
        subout = None

    err_log = False
    if stderr:
        if (stderr & ~NOLOG) == GRAB and not background:
            suberr = subprocess.PIPE
            if stderr == GRAB:
                err_log = True
        elif stderr == STDOUT:
            suberr = subprocess.STDOUT
        elif stderr == NULL:
            if cxlog.is_on():
                suberr = None
            else:
                suberr = open("/dev/null", "w")
        else:
            raise TypeError("invalid value for stderr")
    else:
        suberr = None

    try:
        subp = subprocess.Popen(args, bufsize, executable, None, subout,
                                suberr, preexec_fn, close_fds=True,
                                shell=shell, cwd=cwd, env=env,
                                universal_newlines=universal_newlines)
        if not background:
            out, err = subp.communicate()
            retcode = subp.wait()
        else:
            out = err = ""
            retcode = 0
    except OSError:
        exception = sys.exc_info()[1]
        retcode = -1
        out = ""
        err = cxlog.to_str(exception)

    if isinstance(subout, file):
        # pylint: disable-msg=E0012,E1103
        subout.close()
    if isinstance(suberr, file):
        # pylint: disable-msg=E0012,E1103
        suberr.close()
    if not background and cxlog.is_on():
        cxlog.log("%s%s -> rc=%d  (took %0.3f seconds)" % (logprefix, cxlog.debug_str(os.path.basename(args[0])), retcode, time.time() - start))
        if out_log:
            count = out.count("\n")
            if count > 20:
                cxlog.log("out=<%d lines>" % count)
            else:
                cxlog.log("out=[%s]" % out)
        if err_log:
            count = err.count("\n")
            if count > 20:
                cxlog.log("err=<%d lines>" % count)
            else:
                cxlog.log("err=[%s]" % err)

    return (retcode, out, err)

def system(args, executable=None, preexec_fn=None, shell=False,
           background=False, cwd=None, env=None, logprefix=""):
    """Works the same way as run(), except that neither stdout nor stderr are
    captured, and only the retcode is returned.
    """
    retcode, _out, _err = run(args, executable=executable,
                              preexec_fn=preexec_fn, shell=shell,
                              background=background, cwd=cwd, env=env,
                              logprefix=logprefix)
    return retcode


def launch_url(url):
    """Uses launchurl to open the specified URL in a web browser."""
    return system([os.path.join(CX_ROOT, "bin", "launchurl"), url],
                  background=True)



#####
#
# Select wrapper
#
#####


# Functions and module imports are stored in the same namespace. This means a
# simple 'import select' would get overridden by our select() function, thus
# preventing it from accessing the content of the module, and necessitating
# annoying workarounds (and rightful Pylint warnings).
#
# So we just import the parts we need, renaming them in the process to avoid
# any conflict.
from select import select as select_select
from select import error as select_error

def select(iwtd, owtd, ewtd, timeout=None):
    """Same as select.select() but retries on EINTR (with the same timeout)."""
    while True:
        try:
            return select_select(iwtd, owtd, ewtd, timeout)
        except select_error:
            err = sys.exc_info()[1]
            if err.args[0] == errno.EINTR:
                if timeout is not None:
                    timeout = 0
                continue
            else:
                raise



#####
#
# User directories
#
#####

def _xdg_dirs_skip_whitespace(pos, string):
    while pos < len(string) and string[pos] in ' \t':
        pos += 1
    return pos

def _internal_load_xdg_dirs():
    try:
        # pylint: disable-msg=F0401
        # FIXME: Should we just ship this on Linux? It's small, self-contained, and LGPL.
        import xdg.BaseDirectory
        config_file = xdg.BaseDirectory.load_first_config('user-dirs.dirs')
    except:
        cxlog.warn("Cannot load xdg module:\n%s" % traceback.format_exc())
        config_file = None
    if config_file is None:
        config_file = os.path.join(os.environ.get('HOME', '/'), ".config", "user-dirs.dirs")
    result = {}
    try:
        f = open(config_file, 'U')
    except:
        return {}
    try:
        for line in f:
            line = line.rstrip('\n')
            # Based on MIT code in xdg-user-dir-lookup.c in the xdg-user-dirs package.
            pos = _xdg_dirs_skip_whitespace(0, line)

            if line[pos:pos+4] != 'XDG_':
                continue
            pos += 4

            dir_pos = line.find('_DIR', pos)
            if dir_pos == -1:
                continue
            name = line[pos:dir_pos]
            pos = dir_pos + 4

            pos = _xdg_dirs_skip_whitespace(pos, line)

            if line[pos:pos+1] != '=':
                continue
            pos += 1

            pos = _xdg_dirs_skip_whitespace(pos, line)

            if line[pos:pos+1] != '"':
                continue
            pos += 1

            if line[pos:pos+6] == '$HOME/':
                pos += 6
                value_base = os.environ.get('HOME', '/') + '/'
            elif line[pos:pos+1] == '/':
                value_base = ''
            else:
                continue

            value_end = line.find('"', pos)

            if value_end != -1:
                value = value_base + line[pos:value_end].replace('\\', '')
            else:
                # Yes, the standard parser does not mind if there's no terminating quote.
                value = value_base + line[pos:].replace('\\', '')

            result[name] = value
    finally:
        f.close()
    return result

def _load_xdg_dirs():
    global _xdg_dirs
    try:
        return _xdg_dirs
    except NameError:
        pass
    _xdg_dirs = _internal_load_xdg_dirs()
    return _xdg_dirs

def get_download_dir():
    if distversion.IS_MACOSX:
        # FIXME: OSX-specific code here.
        pass
    else:
        xdg_dirs = _load_xdg_dirs()
        if 'DOWNLOAD' in xdg_dirs:
            return xdg_dirs['DOWNLOAD']
        # FIXME: Localize the default somehow?
    return os.path.join(os.environ.get('HOME', '/'), 'Downloads')

def get_desktop_dir():
    if distversion.IS_MACOSX:
        # FIXME: OSX-specific code here.
        pass
    else:
        xdg_dirs = _load_xdg_dirs()
        if 'DESKTOP' in xdg_dirs:
            return xdg_dirs['DESKTOP']
        # FIXME: Localize the default somehow?
    return os.path.join(os.environ.get('HOME', '/'), 'Desktop')


#####
#
# Locating icons
#
#####


S_MEDIUM = ('48x48', '32x32', '')

def get_icon_path(root, subdir, basename, sizes=S_MEDIUM):
    for ext in ('.png', '.xpm'):
        for size in sizes:
            filename = os.path.join(root, size, subdir, basename + ext)
            if os.path.exists(filename):
                return filename
    cxlog.warn("could not find an icon file for %s %s in %s / %s" % (cxlog.debug_str(basename), sizes, root, subdir))
    return None


#####
#
# pylint utilities
#
#####

def not_yet_implemented():
    """If a method raises NotImplementedError(), pylint regards it as an
    "abstract method" that must be implemented in all subclasses, and the
    class as an "abstract class". Sometimes this is desired, but in some cases
    we use this exception for parts of the API that we are not using now but
    plan to implement and use in the future. Those functions should instead
    raise cxutils.not_yet_implemented()."""
    return NotImplementedError()


#####
#
# umask cache
#
#####

# Retrieve the umask on startup, that is before we have multiple threads all
# trying to get it at the same time. os.umask() MUST NOT be used in anywhere
# else. Should we need to modify the umask, then we will need to write a
# proper umask() wrapper.
UMASK = os.umask(0)
os.umask(UMASK)
